from __future__ import unicode_literals
from six import StringIO, BytesIO

import re
import email
import base64
import quopri
import time
from datetime import datetime
from email.header import decode_header
from imbox.utils import str_encode, str_decode


class Struct(object):
    def __init__(self, **entries):
        self.__dict__.update(entries)

    def keys(self):
        return self.__dict__.keys()

    def __repr__(self):
        return str(self.__dict__)


def decode_mail_header(value, default_charset='us-ascii'):
    """
    Decode a header value into a unicode string.
    """
    try:
        headers = decode_header(value)
    except email.errors.HeaderParseError:
        return str_decode(str_encode(value, default_charset, 'replace'), default_charset)
    else:
        for index, (text, charset) in enumerate(headers):
            try:
                headers[index] = str_decode(text, charset or default_charset,
                                            'replace')
            except LookupError:
                # if the charset is unknown, force default
                headers[index] = str_decode(text, default_charset, 'replace')

        return ''.join(headers)


def get_mail_addresses(message, header_name):
    """
    Retrieve all email addresses from one message header.
    """
    headers = [h for h in message.get_all(header_name, [])]
    addresses = email.utils.getaddresses(headers)

    for index, (address_name, address_email) in enumerate(addresses):
        addresses[index] = {'name': decode_mail_header(address_name),
                            'email': address_email}

    return addresses


def decode_param(param):
    name, v = param.split('=', 1)
    values = v.split('\n')
    value_results = []
    for value in values:
        match = re.search(r'=\?((?:\w|-)+)\?(Q|B)\?(.+)\?=', value)
        if match:
            encoding, type_, code = match.groups()
            try:
                if isinstance(code, str):
                    code = code.encode('ascii')
            except Exception:
                pass
            if type_ == 'Q':
                value = quopri.decodestring(code)
            elif type_ == 'B':
                value = base64.decodebytes(code)
            value = str_encode(value, encoding)
            value_results.append(value)
            if value_results:
                v = ''.join(value_results)
    return name, v


def parse_attachment(message_part):
    # Check again if this is a valid attachment
    content_disposition = message_part.get("Content-Disposition", None)
    if content_disposition is not None:
        dispositions = content_disposition.strip().split(";")

        if dispositions[0].lower() in ["attachment", "inline"]:
            file_data = message_part.get_payload(decode=True)
            if isinstance(file_data, str):
                content = StringIO(file_data)
            else:
                content = BytesIO(file_data)
            if file_data is None:
                return None

            attachment = {
                'content-type': message_part.get_content_type(),
                'size': len(file_data),
                'content': content
            }

            for param in dispositions[1:]:
                name, value = decode_param(param)

                if 'file' in name:
                    attachment['filename'] = value

                if 'create-date' in name:
                    attachment['create-date'] = value

            return attachment

    return None


def decode_content(message):
    try:
        content = message.get_payload(decode=True)
        charset = message.get_content_charset('utf-8')
        if charset != 'utf-8':
            return content.decode(charset)
        return content
    except Exception:
        return 'Could not parse content'


def parse_email(raw_email):
    if isinstance(raw_email, str):
        email_message = email.message_from_string(raw_email)
    else:
        email_message = email.message_from_bytes(raw_email)
    maintype = email_message.get_content_maintype()
    parsed_email = {}

    parsed_email['raw_email'] = raw_email

    body = {
        "plain": [],
        "html": []
    }
    attachments = []

    if maintype == 'multipart':
        for part in email_message.walk():
            content_type = part.get_content_type()
            content_disposition = part.get('Content-Disposition', None)
            if content_disposition:
                content = part.get_payload(decode=True)
            else:
                content = decode_content(part)

            is_inline = content_disposition is None \
                or content_disposition == "inline"
            if content_type == "text/plain" and is_inline:
                body['plain'].append(content)
            elif content_type == "text/html" and is_inline:
                body['html'].append(content)
            elif content_disposition:
                attachment = parse_attachment(part)
                if attachment:
                    attachments.append(attachment)

    elif maintype == 'text':
        payload = decode_content(email_message)
        body['plain'].append(payload)

    parsed_email['attachments'] = attachments

    parsed_email['body'] = body
    email_dict = dict(email_message.items())

    parsed_email['sent_from'] = get_mail_addresses(email_message, 'from')
    parsed_email['sent_to'] = get_mail_addresses(email_message, 'to')

    value_headers_keys = ['subject', 'date', 'message-id']
    key_value_header_keys = ['received-spf',
                             'mime-version',
                             'x-spam-status',
                             'x-spam-score',
                             'content-type']

    parsed_email['headers'] = []
    for key, value in email_dict.items():

        if key.lower() in value_headers_keys:
            valid_key_name = key.lower().replace('-', '_')
            parsed_email[valid_key_name] = decode_mail_header(value)

        if key.lower() in key_value_header_keys:
            parsed_email['headers'].append({'Name': key,
                                            'Value': value})

    if parsed_email.get('date'):
        timetuple = email.utils.parsedate(parsed_email['date'])
        parsed_date = datetime.fromtimestamp(time.mktime(timetuple)) \
            if timetuple else None
        parsed_email['parsed_date'] = parsed_date

    return Struct(**parsed_email)



import datetime
# TODO - Validate query arguments


IMAP_MONTHS = ["Jan", "Feb", "Mar", "Apr", "May", "Jun",
               "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]


def format_date(date):

    return "%s-%s-%s" % (date.day, IMAP_MONTHS[date.month - 1], date.year)


def build_search_query(**kwargs):

    # Parse keyword arguments
    unread = kwargs.get('unread', False)
    sent_from = kwargs.get('sent_from', False)
    sent_to = kwargs.get('sent_to', False)
    date__gt = kwargs.get('date__gt', False)
    if type(date__gt) is datetime.date:
        date__gt = format_date(date__gt)
    date__lt = kwargs.get('date__lt', False)
    if type(date__lt) is datetime.date:
        date__lt = format_date(date__lt)

    query = []

    if unread:
        query.append("(UNSEEN)")

    if sent_from:
        query.append('(FROM "%s")' % sent_from)

    if sent_to:
        query.append('(TO "%s")' % sent_to)

    if date__gt:
        query.append('(SINCE "%s")' % date__gt)

    if date__lt:
        query.append('(BEFORE "%s")' % date__lt)

    if query:
        return " ".join(query)

    return "(ALL)"



from imbox.imap import ImapTransport
from imbox.parser import parse_email
from imbox.query import build_search_query


class Imbox(object):

    def __init__(self, hostname, username=None, password=None, ssl=True):

        self.server = ImapTransport(hostname, ssl=ssl)
        self.username = username
        self.password = password
        self.connection = self.server.connect(username, password)

    def logout(self):
        self.connection.close()
        self.connection.logout()

    def query_uids(self, **kwargs):
        query = build_search_query(**kwargs)

        message, data = self.connection.uid('search', None, query)
        if data[0] is None:
            return []
        return data[0].split()

    def fetch_by_uid(self, uid):
        message, data = self.connection.uid('fetch', uid, '(BODY.PEEK[])')
        raw_email = data[0][1]

        email_object = parse_email(raw_email)

        return email_object

    def fetch_list(self, **kwargs):
        uid_list = self.query_uids(**kwargs)

        for uid in uid_list:
            yield (uid, self.fetch_by_uid(uid))

    def mark_seen(self, uid):
        self.connection.uid('STORE', uid, '+FLAGS', '(\\Seen)')

    def delete(self, uid):
        mov, data = self.connection.uid('STORE', uid, '+FLAGS', '(\\Deleted)')
        self.connection.expunge()

    def copy(self, uid, destination_folder):
        return self.connection.uid('COPY', uid, destination_folder)

    def move(self, uid, destination_folder):
        if self.copy(uid, destination_folder):
            self.delete(uid)

    def messages(self, *args, **kwargs):
        folder = kwargs.get('folder', False)

        if folder:
            self.connection.select(folder)

        return self.fetch_list(**kwargs)

    def folders(self):
        return self.connection.list()



def str_encode(value='', encoding=None, errors='strict'):
    return str(value, encoding, errors)


def str_decode(value, encoding=None, errors='strict'):
    try:
        return bytes(value, encoding, errors).decode('utf-8')
    except:
        try:
            return value.decode(encoding)
        except:
            return str(value)



from setuptools import setup
import os

version = '0.5.6'


def read(filename):
    return open(os.path.join(os.path.dirname(__file__), filename)).read()

setup(
    name='swdt-imbox',
    version=version,
    description="Python IMAP for Human beings. Python3 only",
    long_description=read('README.md'),
    keywords='email, IMAP, parsing emails',
    author='Kirill Bubochkin',
    author_email='ookami.kb@gmail.com',
    url='https://github.com/ookami-kb/imbox',
    license='MIT',
    packages=['imbox'],
    package_dir={'imbox': 'imbox'},
    zip_safe=False,
    install_requires=['six'],
    classifiers=(
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ),
)



from imaplib import IMAP4, IMAP4_SSL


class ImapTransport(object):

    def __init__(self, hostname, port=None, ssl=False):
        self.hostname = hostname
        self.port = port

        if ssl:
            self.transport = IMAP4_SSL
            if not self.port:
                self.port = 993
        else:
            self.transport = IMAP4
            if not self.port:
                self.port = 143

        self.server = self.transport(self.hostname, self.port)

    def list_folders(self):
        return self.server.list()

    def connect(self, username, password):
        self.server.login(username, password)
        self.server.select()

        return self.server
